@@ -3,20 +3,43 @@ from __future__ import print_function
 from collections import Counter, defaultdict, deque
 from functools import partial, wraps
 from heapq import merge
-from itertools import chain, count, islice, takewhile
+from itertools import chain, count, islice, repeat, takewhile, tee
+from operator import itemgetter
 from sys import version_info
 
-from six import iteritems, string_types
+from six import string_types
 from six.moves import filter, map, zip, zip_longest
 
-from .recipes import take
+from .recipes import flatten, take
 
 __all__ = [
-    'bucket', 'chunked', 'collapse', 'collate', 'consumer',
-    'distinct_permutations', 'first', 'ilen', 'interleave_longest',
-    'interleave', 'intersperse', 'iterate', 'one', 'peekable', 'side_effect',
-    'sliced', 'split_after', 'split_before', 'spy', 'unique_to_each',
-    'windowed', 'with_iter'
+    'bucket',
+    'chunked',
+    'collapse',
+    'collate',
+    'consumer',
+    'distinct_permutations',
+    'distribute',
+    'first',
+    'ilen',
+    'interleave_longest',
+    'interleave',
+    'intersperse',
+    'iterate',
+    'one',
+    'padded',
+    'peekable',
+    'side_effect',
+    'sliced',
+    'sort_together',
+    'split_after',
+    'split_before',
+    'spy',
+    'stagger',
+    'unique_to_each',
+    'windowed',
+    'with_iter',
+    'zip_offset',
 ]
 
 
@@ -107,6 +130,10 @@ class peekable(object):
         >>> next(p)
         'b'
 
+    Negative indexes are supported, but be aware that they will cache the
+    remaining items in the source iterator, which may require significant
+    storage.
+
     To test whether there are more items in the iterator, examine the
     peekable's truth value. If it is truthy, there are more items.
 
@@ -166,10 +193,13 @@ class peekable(object):
             ((start is not None) and (start < 0)) or
             ((stop is not None) and (stop < 0))
         ):
-            raise ValueError('Negative indexing not supported')
+            stop = None
+        elif (
+            (start is not None) and (stop is not None) and (start > stop)
+        ):
+            stop = start + 1
 
         cache_len = len(self._cache)
-
         if stop is None:
             self._cache.extend(self._it)
         elif stop >= cache_len:
@@ -181,7 +211,13 @@ class peekable(object):
         if isinstance(index, slice):
             return self._get_slice(index)
 
-        return self._get_slice(slice(index, index + 1, None))[0]
+        cache_len = len(self._cache)
+        if index < 0:
+            self._cache.extend(self._it)
+        elif index >= cache_len:
+            self._cache.extend(islice(self._it, index + 1 - cache_len))
+
+        return self._cache[index]
 
 
 def _collate(*iterables, **kwargs):
@@ -387,48 +423,41 @@ def distinct_permutations(iterable):
 
 
 def intersperse(e, iterable):
-    """Intersperse element ``e`` between the elements of an iterable.
+    """Intersperse element ``e`` between the elements of *iterable*.
 
-        >>> from more_itertools import intersperse
-        >>> list(intersperse('x', [1, 'o', 5, 'k']))
-        [1, 'x', 'o', 'x', 5, 'x', 'k']
+        >>> list(intersperse('x', 'ABCD'))
+        ['A', 'x', 'B', 'x', 'C', 'x', 'D']
         >>> list(intersperse(None, [1, 2, 3]))
         [1, None, 2, None, 3]
-        >>> list(intersperse('x', 1))
-        Traceback (most recent call last):
-        ...
-        TypeError: 'int' object is not iterable
-        >>> list(intersperse('x', []))
-        []
 
     """
     it = iter(iterable)
-    if it:
-        yield next(it)
-        for item in it:
-            yield e
-            yield item
-    raise StopIteration
+    filler = repeat(e)
+    zipped = flatten(zip(filler, it))
+    next(zipped)
+    return zipped
 
 
 def unique_to_each(*iterables):
     """Return the elements from each of the input iterables that aren't in the
     other input iterables.
 
-    For example, suppose packages 1, 2, and 3 have these dependencies:
-        ``pkg_1: (A, B), pkg_2: (B, C), pkg_3: (B, D)``
+    For example, suppose you have a set of packages, each with a set of
+    dependencies::
+
+        {'pkg_1': {'A', 'B'}, 'pkg_2': {'B', 'C'}, 'pkg_3': {'B', 'D'}}
 
     If you remove one package, which dependencies can also be removed?
 
-    If pkg_1 is removed, then A is no longer necessary - it is not associated
-    with pkg_2 or pkg_3. Similarly, C is only needed for pkg_2, and D is
-    only needed for pkg_3:
+    If ``pkg_1`` is removed, then ``A`` is no longer necessary - it is not
+    associated with ``pkg_2`` or ``pkg_3``. Similarly, ``C`` is only needed for
+    ``pkg_2``, and ``D`` is only needed for ``pkg_3``::
 
-        >>> unique_to_each("AB", "BC", "BD")
+        >>> unique_to_each({'A', 'B'}, {'B', 'C'}, {'B', 'D'})
         [['A'], ['C'], ['D']]
 
     If there are duplicates in one input iterable that aren't in the others
-    they will be duplicated in the output. Input order is preserved:
+    they will be duplicated in the output. Input order is preserved::
 
         >>> unique_to_each("mississippi", "missouri")
         [['p', 'p'], ['o', 'u', 'r']]
@@ -442,20 +471,23 @@ def unique_to_each(*iterables):
     return [list(filter(uniques.__contains__, it)) for it in pool]
 
 
-def windowed(seq, n, fillvalue=None):
-    """Return a sliding window (of width n) over data from the iterable.
-
-    When n=2 this is equivalent to ``pairwise(iterable)``.
-    When n is larger than the iterable, ``fillvalue`` is used in place of
-    missing values.
+def windowed(seq, n, fillvalue=None, step=1):
+    """Return a sliding window of width *n* over the given iterable.
 
         >>> all_windows = windowed([1, 2, 3, 4, 5], 3)
-        >>> next(all_windows)
-        (1, 2, 3)
-        >>> next(all_windows)
-        (2, 3, 4)
-        >>> next(all_windows)
-        (3, 4, 5)
+        >>> list(all_windows)
+        [(1, 2, 3), (2, 3, 4), (3, 4, 5)]
+
+    When the window is larger than the iterable, ``fillvalue`` is used in place
+    of missing values::
+
+        >>> list(windowed([1, 2, 3], 4))
+        [(1, 2, 3, None)]
+
+    Each window will advance in increments of *step*::
+
+        >>> list(windowed([1, 2, 3, 4, 5, 6], 3, fillvalue='!', step=2))
+        [(1, 2, 3), (3, 4, 5), (5, 6, '!')]
 
     """
     if n < 0:
@@ -463,6 +495,8 @@ def windowed(seq, n, fillvalue=None):
     if n == 0:
         yield tuple()
         return
+    if step < 1:
+        raise ValueError('step must be >= 1')
 
     it = iter(seq)
     window = deque([], n)
@@ -474,8 +508,18 @@ def windowed(seq, n, fillvalue=None):
     yield tuple(window)
 
     # Appending new items to the right causes old items to fall off the left
+    i = 0
     for item in it:
         append(item)
+        i = (i + 1) % step
+        if i % step == 0:
+            yield tuple(window)
+
+    # If there are items from the iterable in the window, pad with the given
+    # value and emit them.
+    if (i % step) and (step - i < n):
+        for _ in range(step - i):
+            append(fillvalue)
         yield tuple(window)
 
 
@@ -743,3 +787,173 @@ def split_after(iterable, pred):
             buf = []
     if buf:
         yield buf
+
+
+def padded(iterable, fillvalue=None, n=None, next_multiple=False):
+    """Yield the elements from *iterable*, followed by *fillvalue*, such that
+    at least *n* items are emitted.
+
+        >>> list(padded([1, 2, 3], '?', 5))
+        [1, 2, 3, '?', '?']
+
+    If *next_multiple* is ``True``, *fillvalue* will be emitted until the
+    number of items emitted is a multiple of *n*::
+
+        >>> list(padded([1, 2, 3, 4], n=3, next_multiple=True))
+        [1, 2, 3, 4, None, None]
+
+    If *n* is ``None``, *fillvalue* will be emitted indefinitely.
+
+    """
+    it = iter(iterable)
+    if n is None:
+        for item in chain(it, repeat(fillvalue)):
+            yield item
+    elif n < 1:
+        raise ValueError('n must be at least 1')
+    else:
+        item_count = 0
+        for item in it:
+            yield item
+            item_count += 1
+
+        remaining = (n - item_count) % n if next_multiple else n - item_count
+        for _ in range(remaining):
+            yield fillvalue
+
+
+def distribute(n, iterable):
+    """Distribute the items from *iterable* among *n* smaller iterables.
+
+        >>> group_1, group_2 = distribute(2, [1, 2, 3, 4, 5, 6])
+        >>> list(group_1)
+        [1, 3, 5]
+        >>> list(group_2)
+        [2, 4, 6]
+
+    If the length of the iterable is not evenly divisible by n, then the
+    length of the smaller iterables will not be identical::
+
+        >>> children = distribute(3, [1, 2, 3, 4, 5, 6, 7])
+        >>> [list(c) for c in children]
+        [[1, 4, 7], [2, 5], [3, 6]]
+
+    If the length of the iterable is smaller than n, then the last returned
+    iterables will be empty:
+
+        >>> children = distribute(5, [1, 2, 3])
+        >>> [list(c) for c in children]
+        [[1], [2], [3], [], []]
+
+    This function uses ``itertools.tee``, and may require significant storage.
+
+    """
+    if n < 1:
+        raise ValueError('n must be at least 1')
+
+    children = tee(iterable, n)
+    return [islice(it, index, None, n) for index, it in enumerate(children)]
+
+
+def stagger(iterable, offsets=(-1, 0, 1), longest=False, fillvalue=None):
+    """Yield tuples whose elements are offset from *iterable*.
+    The amount by which the ``i``-th item in each tuple is offset is given by
+    the ``i``-th item in *offsets*.
+
+        >>> list(stagger([0, 1, 2, 3]))
+        [(None, 0, 1), (0, 1, 2), (1, 2, 3)]
+        >>> list(stagger(range(8), offsets=(0, 2, 4)))
+        [(0, 2, 4), (1, 3, 5), (2, 4, 6), (3, 5, 7)]
+
+    By default, the sequence will end when the final element of a tuple is the
+    last item in the iterable. To continue until the first element of a tuple
+    is the last item in the iterable, set *longest* to ``True``::
+
+        >>> list(stagger([0, 1, 2, 3], longest=True))
+        [(None, 0, 1), (0, 1, 2), (1, 2, 3), (2, 3, None), (3, None, None)]
+
+    By default, ``None`` will be used to replace offsets beyond the end of the
+    sequence. Specify *fillvalue* to use some other value.
+
+    """
+    children = tee(iterable, len(offsets))
+
+    return zip_offset(
+        *children, offsets=offsets, longest=longest, fillvalue=fillvalue
+    )
+
+
+def zip_offset(*iterables, **kwargs):
+    """``zip`` the input *iterables* together, but offset the ``i``-th iterable
+    by the ``i``-th item in *offsets*.
+
+        >>> list(zip_offset('0123', 'abcdef', offsets=(0, 1)))
+        [('0', 'b'), ('1', 'c'), ('2', 'd'), ('3', 'e')]
+
+    This can be used as a lightweight alternative to SciPy or pandas to analyze
+    data sets in which somes series have a lead or lag relationship.
+
+    By default, the sequence will end when the shortest iterable is exhausted.
+    To continue until the longest iterable is exhausted, set *longest* to
+    ``True``.
+
+        >>> list(zip_offset('0123', 'abcdef', offsets=(0, 1), longest=True))
+        [('0', 'b'), ('1', 'c'), ('2', 'd'), ('3', 'e'), (None, 'f')]
+
+    By default, ``None`` will be used to replace offsets beyond the end of the
+    sequence. Specify *fillvalue* to use some other value.
+
+    """
+    offsets = kwargs['offsets']
+    longest = kwargs.get('longest', False)
+    fillvalue = kwargs.get('fillvalue', None)
+
+    if len(iterables) != len(offsets):
+        raise ValueError("Number of iterables and offsets didn't match")
+
+    staggered = []
+    for it, n in zip(iterables, offsets):
+        if n < 0:
+            staggered.append(chain(repeat(fillvalue, -n), it))
+        elif n > 0:
+            staggered.append(islice(it, n, None))
+        else:
+            staggered.append(it)
+
+    if longest:
+        return zip_longest(*staggered, fillvalue=fillvalue)
+
+    return zip(*staggered)
+
+
+def sort_together(iterables, key_list=(0,), reverse=False):
+    """Return the input iterables sorted together, with *key_list* as the
+    priority for sorting. All iterables are trimmed to the length of the
+    shortest one.
+
+    This can be used like the sorting function in a spreadsheet. If each
+    iterable represents a column of data, the key list determines which
+    columns are used for sorting.
+
+    By default, all iterables are sorted using the ``0``-th iterable::
+
+        >>> iterables = [(4, 3, 2, 1), ('a', 'b', 'c', 'd')]
+        >>> sort_together(iterables)
+        [(1, 2, 3, 4), ('d', 'c', 'b', 'a')]
+
+    Set a different key list to sort according to another iterable.
+    Specifying mutliple keys dictates how ties are broken::
+
+        >>> iterables = [(3, 1, 2), (0, 1, 0), ('c', 'b', 'a')]
+        >>> sort_together(iterables, key_list=(1, 2))
+        [(2, 3, 1), (0, 0, 1), ('a', 'c', 'b')]
+
+    Set *reverse* to ``True`` to sort in descending order.
+
+        >>> sort_together([(1, 2, 3), ('c', 'b', 'a')], reverse=True)
+        [(3, 2, 1), ('a', 'b', 'c')]
+
+    """
+    return list(zip(*sorted(zip(*iterables),
+                            key=itemgetter(*key_list),
+                            reverse=reverse)))
